#!/usr/bin/python3

"""
moiety_modeling.analysis
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module provides several classes to analyze the optimization results, select the optimal model, and visualize the
results. The :class:`~moiety_modeling.analysis.ResultAnalysis` class is responsible for generating general statistics
from the optimization results. The :class:`~moiety_modeling.analysis.ModelRank` class selects the model that best
reflects the observed isotopologue profile. The :class:`~moiety_modeling.analysis.ComparisonTable` class compares
the optimal model selected under different optimization settings. The :class:`~moiety_modeling.analysis.PlotMoietyDistribution`
class plots the distribution of moiety value of the moiety model. The :class:`~moiety_modeling.analysis.PlotIsotopologueIntensity`
class plots comparison of the observed and the calculated isotopologue intensity.

"""


import jsonpickle
import math
import functools
import operator
import statistics as sta
import os
import collections
import csv
import sys
import numpy as np
import matplotlib.pyplot as plt


class ResultsAnalysis:

    """ResultsAnalysis class performs the analysis of moiety model optimization results."""

    def __init__(self, filename, path=None):
        """ResultsAnalysis initializer.

        :param str filename: the filenames of optimization results file.
        :param str path: the path to store the analysis results.
        """
        self.path = path if path is not None else os.path.dirname(filename) + '/'
        with open(filename) as file:
            data = jsonpickle.decode(file.read(), keys=True)
        self.model = data['model']
        self.datasets = data['datasets']
        self.bestGuesses = data['bestGuesses']
        self.optimizationSetting = data['optimizationSetting']
        self.energyFunction = data['energyFunction']
        self.jsonfilePath = None

    def analyze(self):
        """Analyze the optimization results for each model.

        :return dict: the analysis results.
        """
        #calculate the energies.
        energy = [ bestGuess.energy for bestGuess in self.bestGuesses ]

        # calculate the moiety values from best guess.
        moietyValues = []
        # the moietyValues is a list of dictionary. Each dictionary contains the moiety value of a guess, which is also a dictionary. eg[ {datasetName: {}, }, ....]
        for bestGuess in self.bestGuesses:
            moietyValue = {}
            for i in range(len(self.datasets)):
                subVector = [bestGuess.elements[i] for i in range(i* self.model.parameterNum, (i+1) * self.model.parameterNum)]
                moietyValue[self.datasets[i].datasetName] = self.model.calculateMoietyState(subVector)
            moietyValues.append(moietyValue)
        # print(moietyValues[0])

        RSS = []  # this stores the sum of diff square for each optimization result.
        calculatedMolecules = collections.defaultdict(list) # this is a list of calculated molecules for each guess.
        for moietyValue in moietyValues:
            rss = collections.defaultdict(lambda: 0)
            for ds in range(len(self.datasets)):
                dataset = self.datasets[ds]
                dsResidual = 0
                for molecule in self.model.molecules:
                    calculatedMolecule = self._calculateMolecule(molecule, moietyValue[dataset.datasetName])
                    residuals = self._calculateResiduals(molecule, dataset, calculatedMolecule)
                    name = 'DS{0}.{1}'.format(dataset.datasetName, molecule.name)
                    dsResidual += self._calculateRSS(residuals)
                    calculatedMolecules[name].append(calculatedMolecule)
                rss[dataset.datasetName] = dsResidual
            RSS.append(rss)  # one rss for each set of moietyValues separate by the dataset name.

        moietyStats = self._calcualteMoietyStats(moietyValues)

        RSS_AIC = []
        RSS_AICc = []
        BIC = []
        paramsNum = self.model.parameterNum * len(self.datasets)
        dataNum = 0
        for dataset in self.datasets:
            for moleculeName in dataset.keys():
                dataNum += len(dataset[moleculeName])
        for rss in RSS:
            rssSum = sum([rss[key] for key in rss])
            RSS_AIC.append(self._calculateAIC(rssSum, paramsNum, dataNum))
            RSS_AICc.append(self._calculateAICc(rssSum, paramsNum, dataNum))
            BIC.append(self._calculateBIC(rssSum, paramsNum, dataNum))

        optimizeParams = {'AIC': self._calculateMeanStdMaxMin(RSS_AIC), 'AICc': self._calculateMeanStdMaxMin(RSS_AICc), 'BIC': self._calculateMeanStdMaxMin(BIC), 'energy': self._calculateMeanStdMaxMin(energy)}

        # write the descriptive stats results for readability.
        scriptFile = open('{0}{1}_{2}_{3}_Descriptive_Stats.txt'.format(self.path, self.model.name, self.optimizationSetting, self.energyFunction), 'w')
        scriptFile.write("Descriptive Statistics of {0} using {1} and {2} energy function \n\n".format(self.model.name, self.optimizationSetting, self.energyFunction))
        scriptFile.write("Moiety States Statistical Results: \n")

        for dataset in moietyStats:
            scriptFile.write(" Datasets: {0}\n".format(dataset))
            for moietyState in moietyStats[dataset]:
                scriptFile.write("     {0}: {1}\n".format(moietyState, moietyStats[dataset][moietyState]))

        scriptFile.write("\nCalculated Molecule Statistical Results: \n")
        calculatedMoleculesStats = {molecule: {} for molecule in calculatedMolecules.keys()}

        for molecule in calculatedMolecules.keys():
            scriptFile.write(" Calculated {0}:\n".format(molecule))
            # calculateMolecules[molecule] is a list containing results of each optimization. Each item in the list is a dictionary containing all the isotopologues of this molecule.
            for isotopologue in calculatedMolecules[molecule][0].keys():
                isotopologueList = [thisMolecule[isotopologue] for thisMolecule in calculatedMolecules[molecule]]
                stats = self._calculateMeanStdMaxMin(isotopologueList)
                calculatedMoleculesStats[molecule][isotopologue] = stats
                scriptFile.write("     Calculated {0}[{1}]: {2}\n".format(molecule, isotopologue, stats))

        scriptFile.write("\nAverage RSS, RSS_AIC, RSS_AICc, and BIC.\n")
        for key in optimizeParams:
            scriptFile.write(" {0}: {1}\n".format(key, optimizeParams[key]))
        scriptFile.close()

        # write the analysis results into JSON file for reusability.
        self.jsonfilePath = '{0}{1}_{2}_{3}_Stats.json'.format(self.path, self.model.name, self.optimizationSetting, self.energyFunction)
        with open(self.jsonfilePath, 'w') as outJSONfile:
            outJSONfile.write(jsonpickle.encode({'model': self.model.name, 'optimizeParams': optimizeParams, 'moietyValues': moietyValues, 'calculatedMolecules': calculatedMoleculesStats, 'datasets': self.datasets, "optimizationSetting": self.optimizationSetting, 'energyFunction': self.energyFunction}))

        return {'optimizeParams': optimizeParams, 'moietyValues': moietyValues, 'calculatedMolecules': calculatedMoleculesStats}

    def _calculateMolecule(self, molecule, moietyStateValue):

        """To calculate of the energy of the model.

        :param moietyStateValue: the list of value of the moietyStates.
        :return: the energy of the model.
        """

        calculated = collections.defaultdict(lambda: 0)
        for isotopologue in molecule.standardStates:
            calculated[isotopologue] = sum(functools.reduce(operator.mul, [moietyStateValue[i] for i in isotopomer]) for isotopomer in molecule.standardStates[isotopologue])
        return calculated

    def _calculateResiduals(self, molecule, dataset, calculatedMolecules):

        """
        To calculate the residuals of the molecule based on the observed value and observed value.
        :param molecule: the molecule that is checked.
        :param dataset: the dataset that contain the observed value of the molecule.
        :param calculatedMolecules: the calculated values of the molecule.
        :return: the residuals.
        """

        residuals = {}
        observed = dataset[molecule.name] # this is a list
        for isotopologue in observed:
            residuals[isotopologue['labelingIsotopes']] = isotopologue['height'] - calculatedMolecules[isotopologue['labelingIsotopes']]
        return residuals

    def _calculateRSS(self, residuals):

        """
        To calculate the rss of the residuals.
        :param residuals: the residuals
        :return: rss
        """

        return sum([residuals[key] ** 2 for key in residuals.keys()])

    def _calcualteMoietyStats(self, moietyValues):

        """
        To calculate the stats of the moieties for each dataset.
        :param moietyValues: the list of moiety values. [{datasetName: {a[0]: 0.01}, }, {}, ]
        :return: the dictionary of moiety value.
        """

        averagedParams = {dataset.datasetName: {} for dataset in self.datasets}
        for dataset in self.datasets:
            for moiety in moietyValues[0][dataset.datasetName].keys():
                moietyValueList= [moietyValue[dataset.datasetName][moiety] for moietyValue in moietyValues]
                averagedParams[dataset.datasetName][moiety] = self._calculateMeanStdMaxMin(moietyValueList)
        return averagedParams

    def _calculateMeanStdMaxMin(self, list):

        """
        To calculate the mean, std, max and min of the list.
        :param list: a list of value.
        :return: dictionary of count, mean, std, max and min.
        """

        count = len(list)
        mean = sum(list)/count
        std = sta.stdev(list)
        maxValue = max(list)
        minValue = min(list)
        stat = {'count': count, 'mean': mean, 'std': std, 'max': maxValue, 'min': minValue}
        return stat

    def _calculateAIC(self, rss, paramsNum, dataNum):

        """
        To calculate the Akaike Information Criterion.
        :param rss: residual sum of squares.
        :param paramsNum: number of parameters (degrees of freedom).
        :param dataNum: number of data points.
        :return: the AIC value.
        """

        return 2 * paramsNum + dataNum * math.log(rss/dataNum)

    def _calculateAICc(self, rss, paramsNum, dataNum):

        """
        Calculate the corrected Akaike Information Criterion. AICc is useful when dataNum is not many times more than paramsNum^2.
        :param rss: residual sum of squares.
        :param paramsNum: number of parameters (degrees of freedom).
        :param dataNum: number of data points.
        :return: the AICc value.
        """

        return self._calculateAIC(rss, paramsNum, dataNum) + (2 * paramsNum * (paramsNum + 1)) / (dataNum - paramsNum -1)

    def _calculateBIC(self, rss, paramsNum, dataNum):

        """
        To calculate the Bayesian Information Criterion.
        :param rss: residual sum of square.
        :param paramsNum: number of parameters (degrees of freedom).
        :param dataNum: number of data points.
        :return: the BIC value.
        """

        return dataNum * math.log(rss/dataNum) + paramsNum * math.log(dataNum)


class ModelRank:

    """ModelRank class ranks the models according to the selection criteria."""

    def __init__(self, pathFile, path, selectionCriterion):
        """ModelRank initializer.

        :param str pathFile: the txt file containing paths to the model analysis results.
        :param str path: the path to store the model rank results.
        :param str selectionCriterion: the selection criteria (eg: AIC, BIC, BICc).
        """
        self.selectionCriterion = selectionCriterion
        self.path = path if path is not None else '{0}/model_rank_{1}/'.format(os.path.dirname(pathFile), self.selectionCriterion)
        if not os.path.exists(self.path):
            os.mkdir(self.path)
        self.pathFile = pathFile

    def rank(self):
        """To rank the models according to the selection criteria.

        :return list: the rank results.
        """

        dataCollection = {}
        energyFunction = None
        optimizationSetting = None

        with open(self.pathFile) as pathFile:
            for filename in pathFile:
                filename = filename.replace("\n", "")
                with open(filename) as file:
                    data = jsonpickle.decode(file.read())
                    # keys are the AIC, AICc, BIC and energy.
                    dataCollection[data['model']] = data['optimizeParams'][self.selectionCriterion]['mean']
                    if energyFunction is None:
                        energyFunction = data['energyFunction']
                    else:
                        if energyFunction != data['energyFunction']:
                            sys.exit("Models cannot be compared with different energyFunction!")
                    if optimizationSetting is None:
                        optimizationSetting = data['optimizationSetting']
                    else:
                        if optimizationSetting != data['optimizationSetting']:
                            sys.exit("Models cannot be compared with different optimizationSetting!")

        # the sorted function return the ordered items in the dictionary, each item is a set of key and value, like (key, value).
        rankedData = sorted(dataCollection.items(), key=operator.itemgetter(1))

        with open('{0}Model_rank_on_{1}.json'.format(self.path, self.selectionCriterion), 'w') as outFile:
            outFile.write(jsonpickle.encode({'rank': rankedData, 'selectionCriteria': self.selectionCriterion, 'optimizationSetting': optimizationSetting, 'energyFunction': energyFunction}))

        with open('{0}Model_rank_on_{1}.txt'.format(self.path, self.selectionCriterion), 'w') as outFile:
            outFile.write("Model rank of {0} based on model selection criteria.\n".format(self.selectionCriterion))
            for data in rankedData:
                outFile.write('{0}: optimizeParams: {1} \n'.format(data[0], data[1]))

        return rankedData


class ComparisonTable:

    """
    ComparisonTable class collects the best and second best models under different optimization settings.
    """

    def __init__(self, pathFile, path):
        """ComparisonTable initializer.

        :param str pathFile: the filenames of model rank at different situations.
        :param str path: the path to directory that stores the result.
        """

        self.path = path if path is not None else os.path.dirname(pathFile) + '/'
        self.pathFile = pathFile

    def makeTable(self):

        """To make the comparison table under different optimization settings.

        :return: None
        :rtype: :py:obj:`None`
        """

        dataCollection = {}
        energyFunction = None
        selectionCriteria = None

        with open(self.pathFile) as pathFile:
            for filename in pathFile:
                filename = filename.replace("\n", "")
                with open(filename) as file:
                    data = jsonpickle.decode(file.read())
                    dataCollection[data['optimizationSetting']] = [data['rank'][i] for i in range(2)]
                    if energyFunction is None:
                        energyFunction = data['energyFunction']
                    else:
                        if energyFunction != data['energyFunction']:
                            sys.exit("The analysis results have different energy function!")
                    if selectionCriteria is None:
                        selectionCriteria = data['selectionCriteria']
                    else:
                        if selectionCriteria != data['selectionCriteria']:
                            sys.exit("The analysis results have different selection criteria!")

        with open('{0}Comparison_table_of_optimization_parameter_based_on_{1}_with_{2}_energy_function.csv'.format(self.path, selectionCriteria, energyFunction), 'w') as csvFile:
            csvWriter = csv.writer(csvFile, delimiter=',')
            csvWriter.writerow('Energy function: {0}'.format(energyFunction))
            csvWriter.writerow('Selection criteria: {0}'.format(selectionCriteria))
            csvWriter.writerow(['optimizationParameter', '1st model', selectionCriteria, '2nd model', selectionCriteria, 'abs_diff', 'relative_diff'])
            for rank in dataCollection:
                firstModel = dataCollection[rank][0][0]
                secondModel = dataCollection[rank][1][0]
                diff = abs(dataCollection[rank][0][1] - dataCollection[rank][1][1])
                relativeDiff = diff / abs(dataCollection[rank][1][1])
                csvWriter.writerow([rank, firstModel, dataCollection[rank][0][1], secondModel, dataCollection[rank][1][1], diff, relativeDiff])


class PlotMoietyDistribution:

    """
    PlotMoietyDistribution class plots the moiety state distribution of the optimization results.
    """

    def __init__(self, filename, path):
        """PlotMoietyDistribution initializer.

        :param str filename: the json file of model analysis results.
        :param str path: the path to store the plot results.
        """
        self.filename = filename
        self.path = path

    def plotMoiety(self):
        """To plot the distribution of moiety states.

        :return: None
        :rtype: :py:obj:`None`
        """
        with open(self.filename) as file:
            data = jsonpickle.decode(file.read())
            model = data['model']
            optimizationSetting = data['optimizationSetting']
            energyFunction = data['energyFunction']
            moietyValues = data['moietyValues']
            moietyCollections = {}
            for dataset in moietyValues[0].keys():
                for moietyState in moietyValues[0][dataset].keys():
                    moietyValueList = [moietyValue[dataset][moietyState] for moietyValue in moietyValues]
                    moietyCollections['DS{0}.{1}'.format(dataset, moietyState)] = moietyValueList
            nPlots = len(moietyCollections)
            labels = sorted(moietyCollections.keys())
            for i in range(nPlots):
                value = moietyCollections[labels[i]]
                plt.subplot(int(nPlots/5)+1, 5, i+1)
                plt.hist(value, 30, range=[0,1], align='mid', alpha=0.5, density=True)
                plt.xlabel("Moiety Value")
                plt.ylabel("Counts")
                plt.title("{0}".format(labels[i]), fontsize=10)
            plt.suptitle('Moiety value histogram of model {0} with {1} optimization setting and {2} energy function'.format(model, optimizationSetting, energyFunction))
            plt.tight_layout()
            plt.show()


class PlotIsotopologueIntensity:

    """
    PlotIsotopologueIntensity class plots the comparison of calculated and observed isotoplogue intensity.
    """

    def __init__(self, filename, path):
        """PlotIsotopologueIntensity initializer.

        :param str filename: The json file of model analysis results.
        :param str path: the path to store the plot results.
        """
        self.filename = filename
        self.path = path

    def plotIsotopologue(self):
        """To plot the comparison of observed and calculated isotoplogue intensities.

        :return: None
        :rtype: :py:obj:`None`
        """
        with open(self.filename) as file:
            data = jsonpickle.decode(file.read())
            calculatedMolecules = data['calculatedMolecules']
            observedMolecules = data['datasets']
        molecules = calculatedMolecules.keys()
        calValue = {molecule :{} for molecule in molecules}
        calStd = {molecule: {} for molecule in molecules}
        obsValue = {molecule: {} for molecule in molecules}
        obsStd = {molecule: {} for molecule in molecules}
        for molecule in molecules:
            for isotopologue in calculatedMolecules[molecule].keys():
                calValue[molecule][isotopologue] = calculatedMolecules[molecule][isotopologue]['mean']
                calStd[molecule][isotopologue] = calculatedMolecules[molecule][isotopologue]['std']

        for dataset in observedMolecules:

            for molecule in dataset:
                name = 'DS{0}.{1}'.format(dataset.datasetName, molecule)
                if name in molecules:
                    for isotopologue in dataset[molecule]:
                        obsValue[name][isotopologue['labelingIsotopes']] = isotopologue['height']
                        obsStd[name][isotopologue['labelingIsotopes']] = isotopologue['heightSE']

        for molecule in molecules:
            labels = sorted(calValue[molecule].keys())
            calPerformance = [calValue[molecule][label] for label in labels]
            calMoleculeSta = [calStd[molecule][label] for label in labels]
            obsPerformance = [obsValue[molecule][label] for label in labels]
            obsMoleculeSta = [obsStd[molecule][label] for label in labels]
            fig, ax = plt.subplots()
            index = np.arange(len(labels))
            barWidth = 0.2
            opacity = 0.4
            errorConfig = {"ecolor": '0.3'}
            rects1 = plt.bar(index, calPerformance, barWidth, alpha=opacity, color='b', yerr=calMoleculeSta, error_kw=errorConfig, label='calculated' )
            rects2 = plt.bar(index+barWidth, obsPerformance, barWidth, alpha=opacity, color='r', yerr=obsMoleculeSta, error_kw=errorConfig, label='observed' )
            plt.xlabel('Isotopologues of {0}'.format(molecule))
            plt.ylabel('Normalized Intensity')
            plt.xticks(index+barWidth/2, labels)
            plt.legend()
            plt.tight_layout()
            plt.show()













#!/usr/bin/python3

"""
moiety_modeling.modeling
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module provides the :class:`~moiety_modeling.modeling.Dataset` class to organize a single mass spectroscopy isotopologue
profile dataset into a dictionary-based data structure, the :class:`~moiety_modeling.modeling.ModelOptimization` and derived
classes for performing a single model optimization, and the :class:`~moiety_modeling.modeling.OptimizationManager` class
to manage the optimization process of moiety modeling.

The optimization results are stored in the JSON file. A txt file containing all the paths to the generated ``JSON`` files is
created to facilitate further model analysis.
"""

import sys
import SAGA_optimize
import operator
import functools
import jsonpickle
import collections
import multiprocessing
import os
import scipy
import scipy.optimize
import abc
import random
import math
import logging
import shutil

__all__ = ['Dataset', 'OptimizationManager', 'ModelOptimization', 'SAGAoptimization', 'SAGAseparateOptimization', 'ScipyGuess', 'ScipyOptimization', 'ScipySeparateOptimization']


class Dataset(collections.UserDict):

    """Dataset class that stores a single mass spectroscopy isotopologue profile in the form of :py:class:`~collections.UserDict`."""

    def __init__(self, datasetName, *args, **kwargs):
        """Dataset initializer.

        :param str datasetName: the name of the dataset.
        """
        self.datasetName = datasetName
        super().__init__(*args, **kwargs)


class ModelOptimization(abc.ABC):

    """
    The abstract :class:`~moiety_modeling.modeling.ModelOptimization` class.
    """

    def __init__(self, model, datasets, path, methodParameters, optimizationSetting, energyFunction):

        """ModelOptimization initializer.

        :param model: a :class:`~moiety_modeling.model.Model` instance.
        :type  model: :class:`~moiety_modeling.model.Model`
        :param list datasets: a list of :class:`~moiety_modeling.modeling.Dataset` instances.
        :param str path: the subdirectory path to save the optimization results.
        :param dict methodParameters: the parameters for optimization method.
        :param str optimizationSetting: abbreviated name for the optimization.
        :param str energyFunction: the energy function used in the optimization.
        """

        self.model = model
        self.datasets = datasets
        self.path = path
        self.methodParameters = methodParameters
        self.optimizationSetting = optimizationSetting
        self.energyFunction = energyFunction
        self.functionCollection = {'absDifference': self.absDifferenceEnergyFunction, 'logDifference': self.logDifferenceEnergyFunction}
        self.activeEnergyFunction = self.functionCollection[self.energyFunction]
        self.bestGuesses = []
        self.elements = self._createElements()

    def _createElements(self):

        """
        To create parameters for optimization.
        :return: a list of parameters.
        """

        elements = []
        for dataset in self.datasets:
            for moiety in self.model.orderedMoieties:
                elements += [ '{0}.{1}'.format(dataset.datasetName, element) for element in self.model.properties[moiety.name]['elementNames'] ]
        return elements

    def energyCalculation(self, vector):

        """Calculate the energy of the model.

        :param list vector: a list of model parameter values.
        :return double: the energy.
        """

        energy = 0
        for i in range(len(self.datasets)):
            subVector = [ vector[i] for i in range(i* self.model.parameterNum, (i+1) * self.model.parameterNum) ]
            energy += self.activeEnergyFunction(self.model.calculateMoietyState(subVector), self.datasets[i])
        return energy

    def absDifferenceEnergyFunction(self, moietyStateValue, dataset):

        """The absolute difference energy function. The absolute value between the observed and the calculated isotopologues: energy = sum(|I<cal> - I<obs>|).

        :param list moietyStateValue: a list of values for the moietyStates.
        :param dataset: a :class:`~moiety_modeling.modeling.Dataset` instance.
        :type dataset: :class:`~moiety_modeling.modeling.Dataset`
        :return: the energy of the model.
        :rtype: double
        """

        energy = 0
        for molecule in self.model.molecules:
            if len(dataset[molecule.name]) != len(molecule.allStates):
                sys.exit("The isotologue data of {0} is not enough.".format(molecule.name))
            calculated = collections.defaultdict(lambda: 0)
            for isotopologue in molecule.standardStates:
                calculated[isotopologue] = sum(functools.reduce(operator.mul, [moietyStateValue[i] for i in isotopomer]) for isotopomer in molecule.standardStates[isotopologue])
            energy += sum([abs(calculated[isotopologue['labelingIsotopes']] - isotopologue['height']) for isotopologue in dataset[molecule.name]])
        return energy

    def logDifferenceEnergyFunction(self, moietyStateValue, dataset):

        """The log difference energy function. The difference between the log of observed and the calculated isotopologues: energy = sum(|log(I<cal>) - log(I<obs>)|)

        :param list moietyStateValue: a list of values for the moietyStates.
        :param dataset: a :class:`~moiety_modeling.modeling.Dataset` instance.
        :type dataset: :class:`~moiety_modeling.modeling.Dataset`
        :return: the energy of the model.
        :rtype: double
        """

        energy = 0
        for molecule in self.model.molecules:
            if len(dataset[molecule.name]) != len(molecule.allStates):
                sys.exit("The isotopologue data of {0} is not enough.".format(molecule.name))
            calculated = collections.defaultdict(lambda: 0)
            for isotopologue in molecule.standardStates:
                calculated[isotopologue] = sum(functools.reduce(operator.mul, [moietyStateValue[i] for i in isotopomer]) for isotopomer in molecule.standardStates[isotopologue])
            leastIsotopologue = min([isotopologue['height'] for isotopologue in dataset[molecule.name] if isotopologue['height'] > 0]) / 2
            for isotopologue in dataset[molecule.name]:
                if calculated[isotopologue['labelingIsotopes']] != 0 and isotopologue['height'] == 0:
                    energy += abs(math.log(calculated[isotopologue['labelingIsotopes']]) - math.log(leastIsotopologue))
                elif calculated[isotopologue['labelingIsotopes']] == 0 and isotopologue['height'] != 0:
                    energy += abs(math.log(leastIsotopologue) - math.log(isotopologue['height']))
                elif calculated[isotopologue['labelingIsotopes']] != 0 and isotopologue['height'] != 0:
                    energy += abs(math.log(calculated[isotopologue['labelingIsotopes']]) - math.log(isotopologue['height']))
        return energy

    def optimizationScripts(self):

        """
        To save the optimization scripts of the optimization for check.
        :return: None
        :rtype: :py:obj:`None`
        """

        outputFile = open('{0}{1}_{2}_optimization_scripts.txt'.format(self.path, self.model.name, self.optimizationSetting), 'w')
        outputFile.write(self.model.name+'\n')
        for moiety in self.model.moieties:
            outputFile.write(str(moiety)+'\n')
        for relationship in self.model.relationships:
            outputFile.write(str(relationship)+'\n')
        for molecule in self.model.molecules:
            outputFile.write(str(molecule)+'\n')
        for dataset in self.datasets:
            outputFile.write("Dataset {0}\n".format(dataset.datasetName))
            for molecule in self.model.molecules:
                outputFile.write(" {0} data = {1}\n".format(molecule.name, dataset[molecule.name]))
        outputFile.write(str(self.model))
        string = "\n energy = 0\n"
        for molecule in self.model.molecules:
            string += " {0}_cal = {1}\n".format(molecule.name, {labelingIsotopes: 0 for labelingIsotopes in molecule.allStates})
            for isotopologue in molecule.allStates:
                string += " {0}_calc[{1}] = {2}\n".format(molecule.name, isotopologue, ' + '.join(['.'.join(isotopomer) for isotopomer in molecule.standardStates[isotopologue]]))
            string += "\n {0}Obeserved = data['{1}']\n".format(molecule.name, molecule.name)
            string += " for x in range(len({0}Observed)):\n".format(molecule.name)
            if self.energyFunction == 'absDifference':
                string += "     energy += abs({0}_calc[{0}Observed[x]['labelingIsotopes']] - {0}Observed[x]['height'])\n".format(molecule.name)
            elif self.energyFunction == 'logDifference':
                string += "     energy += abs(log({0}_calc[{0}Observed[x]['labelingIsotopes']) - log({0}Observed[x]['height']\n)".format(molecule.name)
            elif self.energyFunction == 'proportionalDifference':
                string += "     energy += abs(({0}_calc[{0}Observed[x]['labelingIsotopes']] - {0}Observed[x]['height']) / {0}Observed[x]['height'])\n".format(molecule.name)
        string += ' return energy'
        outputFile.write(string)
        outputFile.close()


class SAGAoptimization(ModelOptimization):

    """
    The :class:`~moiety_modeling.modeling.SAGAoptimization` class (for combined datasets) inherited from :class:`~moiety_modeling.modeling.ModelOptimization`.
    """

    def __init__(self, model, datasets, path, methodParameters, optimizationSetting, energyFunction, noPrintBestResults, noPrintAllResults):

        """SAGAoptimization initializer.

        :param model: a :class:`~moiety_modeling.model.Model` instance.
        :type  model: :class:`~moiety_modeling.model.Model`
        :param list datasets: a list of :class:`~moiety_modeling.modeling.Dataset` instances.
        :param str path: the path to save the optimization results.
        :param dict methodParameters: the parameters for optimization method.
        :param str optimizationSetting: abbreviated name for the optimization.
        :param str energyFunction: the energy function used in the optimization.
        :param int noPrintBestResults: not to save the all the best results of the optimization process.
        :param int noPrintAllResults: not to save the all the results of the optimization process.
        """
        super().__init__(model, datasets, path, methodParameters, optimizationSetting, energyFunction)
        self.noPrintBestResults = noPrintBestResults
        self.noPrintAllResults = noPrintAllResults

    def bestResultsFile(self, i):

        """Open the file to record the best results during the optimization process.

        :param int i: the ith optimization.
        :return: the file handler.
        :rtype: :py:class:`~_io.TextIOWrapper`
        """

        if self.noPrintBestResults:
            return None

        return open('{0}{1}_best/{1}_best_results_{2}_{3}.txt'.format(self.path, self.model.name, self.optimizationSetting, i), 'w')

    def allResultsFile(self, i):

        """Open the file to record the all optimization results.

        :param int i: the ith optimization.
        :return: the file handler.
        :rtype: :py:class:`~_io.TextIOWrapper`.
        """

        if self.noPrintAllResults:
            return None

        return open('{0}{1}_all/{1}_best_results_{2}_{3}.txt'.format(self.path, self.model.name, self.optimizationSetting, i), 'w')

    def optimizeSingle(self, i):

        """To perform one optimization.

        :param int i: the ith optimization.
        :return: the best :class:`~SAGA_optimize.Guess` from the optimization process.
        :rtype: :class:`~SAGA_optimize.Guess`.
        """

        saga = SAGA_optimize.SAGA(**self.methodParameters, bestResultsFile=self.bestResultsFile(i), allResultsFile=self.allResultsFile(i), energyCalculation=self.energyCalculation)
        for element in self.elements:
            saga.addElementDescriptions(SAGA_optimize.ElementDescription(name=element, low=0, high=1))
        optimizedPopulation = saga.optimize()
        if saga.allResultsFile:
            saga.allResultsFile.close()
        if saga.bestResultsFile:
            saga.bestResultsFile.close()
        return optimizedPopulation.bestGuess

    def creatSubdir(self):

        """To create subdirectories to store the optimization process.

        :return: None.
        :rtype: :py:obj:`None`.
        """

        if not self.noPrintBestResults:
            os.makedirs(self.path+self.model.name+'_best')
        if not self.noPrintAllResults:
            os.makedirs(self.path + self.model.name + '_all')


class SAGAseparateOptimization(SAGAoptimization):

    """
    The :class:`~moiety_modeling.modeling.SAGAseparateOptimization` class (for split dataset) inherited from :class:`~moiety_modeling.modeling.SAGAoptimization`.
    """

    def __init__(self, model, datasets, path, methodParameters, optimizationSetting, energyFunction, noPrintBestResults, noPrintAllResults):

        """SAGAseparateOptimization initializer.

        :param model: a :class:`~moiety_modeling.model.Model` instance.
        :type  model: :class:`~moiety_modeling.model.Model`
        :param list datasets: a list of :class:`~moiety_modeling.modeling.Dataset` instances.
        :param str path: the path to save the optimization results.
        :param dict methodParameters: the parameters for optimization method.
        :param str optimizationSetting: abbreviated name for the optimization.
        :param str energyFunction: the energy function used in the optimization.
        :param int noPrintBestResults: not to save the all the best results of the optimization process.
        :param int noPrintAllResults: not to save the all the results of the optimization process.
        """
        super().__init__(model, datasets, path, methodParameters, optimizationSetting, energyFunction, noPrintBestResults, noPrintAllResults)

    def optimizeSingle(self, i):

        """Perform one optimization.

        :param int i: the ith optimization.
        :return: the best :class:`~SAGA_optimize.Guess` from the optimization process.
        :rtype: :class:`~SAGA_optimize.Guess`.
        """

        bestGuesses = []
        for dataset in self.datasets:
            optimization = SAGAoptimization(self.model, [dataset], self.path, self.methodParameters, self.optimizationSetting + '_' + dataset.datasetName, self.energyFunction, self.noPrintBestResults, self.noPrintAllResults)
            bestGuesses.append(optimization.optimizeSingle(i))
        combinedGuess = bestGuesses[0]
        for i in range(1, len(bestGuesses)):
            combinedGuess.elementDescriptions += bestGuesses[i].elementDescriptions
            combinedGuess.elements += bestGuesses[i].elements
            combinedGuess.energy += bestGuesses[i].energy
        return combinedGuess


class ScipyGuess:
    """
    To convert optimization results to :class:`~SAGA_optimize.Guess` instance.
    """

    def __init__(self, elements, energy):

        """ScipyGuess initializer.

        :param list elements: a list of values for model parameters.
        :param double energy: the energy of the Guess calculated from an energy function.
        """

        self.elements = elements
        self.energy = energy


class ScipyOptimization(ModelOptimization):

    """
    The :class:`~moiety_modeling.modeling.ScipyOptimization` class (for combined datasets) inherited from :class:`~moiety_modeling.modeling.ModelOptimization`.
    """

    def __init__(self, model, datasets, path, methodParameters, optimizationSetting, energyFunction, method):

        """ScipyOptimization initializer.

        :param model: the :class:`~moiety_modeling.model.Model` instance.
        :type model: :class:`~moiety_modeling.model.Model`
        :param list datasets: a list of :class:`~moiety_modeling.modeling.Dataset` instances.
        :param str path: the path to save the optimization results.
        :param dict methodParameters: the parameters for optimization method.
        :param str optimizationSetting: abbreviated name for the optimization.
        :param str energyFunction: the energy function used in the optimization.
        :param str method: the scipy optimization method.
        """

        super().__init__(model, datasets, path, methodParameters, optimizationSetting, energyFunction)
        self.method = method

    def optimizeSingle(self, i):

        """To perform one optimization.

        :param i: the ith optimization.
        :return: the best :class:`~moiety_modeling.modeling.ScipyGuess` from the optimization process.
        :rtype: :class:`~moiety_modeling.modeling.ScipyGuess`.
        """

        bounds = scipy.array([(0, 1) for _ in self.elements])
        fun = lambda vector: self.energyCalculation(vector)
        while True:
            initialGuess = scipy.array([random.random() for _ in self.elements])
            optimizeResult = scipy.optimize.minimize(fun, initialGuess, method=self.method, bounds=bounds, options=self.methodParameters)
            if optimizeResult.success:
                bestGuess = ScipyGuess(optimizeResult.x.tolist(), fun(optimizeResult.x.tolist()))
                break
            else:
                continue
        return bestGuess


class ScipySeparateOptimization(ScipyOptimization):

    """
    The :class:`~moiety_modeling.modeling.ScipySeparateOptimization` class (for split dataset) inherited from :class:`~moiety_modeling.modeling.ScipyOptimization`.
    """

    def __init__(self, model, datasets, path, methodParameters, energyFunction, optimizationSetting, method):

        """ScipySeparateOptimization initializer.

        :param model: the :class:`~moiety_modeling.model.Model` instance.
        :type model: :class:`~moiety_modeling.model.Model`
        :param list datasets: a list of :class:`~moiety_modeling.modeling.Dataset` instances.
        :param str path: the path to save the optimization results.
        :param dict methodParameters: the parameters for optimization method.
        :param str energyFunction: the energy function used in the optimization.
        :param str optimizationSetting: abbreviated name for the optimization.
        :param str method: the scipy optimization method.
        """
        super().__init__(model, datasets, path, methodParameters, energyFunction, optimizationSetting, method)

    def optimizeSingle(self, i):

        """To perform single optimization.

        :param int i: the ith optimization.
        :return: the best :class:`~moiety_modeling.modeling.ScipyGuess` from the optimization process.
        :rtype: :class:`~moiety_modeling.modeling.ScipyGuess`.
        """

        bestGuesses = []
        for dataset in self.datasets:
            optimization = ScipyOptimization(self.model, [dataset], self.path, self.methodParameters, self.optimizationSetting + '_' + dataset.datasetName, self.energyFunction, self.method)
            bestGuesses.append(optimization.optimizeSingle(i))
        elements = []
        energy = 0
        for i in range(len(bestGuesses)):
            elements += bestGuesses[i].elements
            energy += bestGuesses[i].energy
        return ScipyGuess(elements, energy)


class OptimizationManager:

    """OptimizationManager class manage the optimization process based on the optimization settings."""

    def __init__(self, models, datasets, optimizations, path, split=True, multiprocess=True, force=True, times=100, energyFunction='logDifference', printOptimizationScript=True):
        """OptimizationManager initializer.

        :param list models: a list of :class:`~moiety_modeling.model.Moiety` instances.
        :param list datasets: a list of :class:`~moiety_modeling.modeling.Dataset` instances.
        :param dict optimizations: a list of optimization settings.
        :param str path: the path to store the optimization results.
        :param True split: to split the dataset or not.
        :param True multiprocess: to apply multiprocess or not.
        :param True force: to force optimization process or not.
        :param int times: the times of optimization for each moiety model.
        :param str energyFunction: the energy function used for optimization.
        :param True printOptimizationScript: to print out the optimization script or not.
        """

        self.models = models
        self.datasets = datasets
        self.optimizations = optimizations
        self.path = path
        self.split = split
        self.multiprocess = multiprocess
        self.force = force
        self.times = times
        self.energyFunction = energyFunction
        self.printOptimizationScript = printOptimizationScript

    def optimizeModels(self):
        """To optimize moiety models based on the optimization settings.

        :return: None
        :rtype: :py:obj:`None`
        """

        logger = self._initLogging()

        for optimizationMethod in self.optimizations:

            for setting in self.optimizations[optimizationMethod]:

                optimizationParameters = setting

                if not os.path.exists(self.path + '/' + optimizationParameters['optimizationSetting']):
                    os.mkdir(self.path + '/' + optimizationParameters['optimizationSetting'])
                path = self.path + '/' + optimizationParameters['optimizationSetting'] + '/'

                for model in self.models:

                    datasets = []
                    datasetName = ''

                    for dataset in self.datasets:
                        # datasets can contain more molecules than model.

                        if set([molecule.name for molecule in model.molecules]).issubset(set(dataset.keys())):
                            datasetName += dataset.datasetName + " "
                            datasets.append(dataset)

                    if datasets:
                        logger.info("Performing {0} optimization on {1} with {2}dataset".format(optimizationParameters['optimizationSetting'], model.name, datasetName))

                        if optimizationMethod == 'SAGA':
                            if self.split:
                                optimization = SAGAseparateOptimization(model, datasets, path,  optimizationParameters['methodParameters'], optimizationParameters['optimizationSetting'], self.energyFunction, optimizationParameters['noPrintBestResults'], optimizationParameters['noPrintAllResults'])
                            else:
                                optimization = SAGAoptimization(model, datasets, path, optimizationParameters['methodParameters'], optimizationParameters['optimizationSetting'], self.energyFunction, optimizationParameters['noPrintBestResults'], optimizationParameters['noPrintAllResults'])
                            optimization.creatSubdir()
                        elif optimizationMethod in ['L-BFGS-B', 'TNC', 'SLSQP']:
                            if self.split:
                                optimization = ScipySeparateOptimization(model, datasets, path, optimizationParameters['methodParameters'], optimizationParameters['optimizationSetting'], self.energyFunction, optimizationMethod)
                            else:
                                optimization = ScipyOptimization(model, datasets, path, optimizationParameters['methodParameters'], optimizationParameters['optimizationSetting'], self.energyFunction, optimizationMethod)
                        else:
                            logger.warning("The optimization optimizationMethod does not exist for {0} with {1}.".format(model.name, optimizationParameters['optimizationSetting']))
                            if self.force:
                                continue
                            else:
                                sys.exit("Optimization stops with error.")

                        if self.multiprocess:
                            try:
                                with multiprocessing.Pool() as pool:
                                    optimization.bestGuesses = pool.map(optimization.optimizeSingle, (i + 1 for i in range(self.times)))
                            except Exception:
                                logger.exception("{0} with {1} optimization setting fails at multiprocessing".format(model.name, optimizationParameters['optimizationSetting']))
                                if self.force:
                                    continue
                                else:
                                    sys.exit("Optimization stops with error.")
                        else:
                            for i in range(self.times):
                                try:
                                    optimization.bestGuesses.append(optimization.optimizeSingle(i))
                                except Exception:
                                    logger.exception("{0} with {1} optimization setting fails at {2} iteration".format(model.name, optimizationParameters['optimizationSetting'], i))
                                    if self.force:
                                        continue
                                    else:
                                        sys.exit("Optimization stops with error.")

                        # to compress the SAGA optimization results
                        if optimizationMethod == 'SAGA':
                            if os.path.exists(optimization.path + model.name + '_all'):
                                shutil.make_archive(optimization.path + model.name + '_all', 'zip', optimization.path + model.name + '_all')
                                shutil.rmtree(optimization.path + model.name + '_all')
                            if os.path.exists(optimization.path + model.name +'_best'):
                                shutil.make_archive(optimization.path + model.name + '_best', 'zip', optimization.path + model.name + '_best')
                                shutil.rmtree(optimization.path + model.name + '_best')

                        if self.printOptimizationScript:
                            optimization.optimizationScripts()

                        jsonpickle.set_encoder_options('json', sort_keys=True, indent=4)
                        with open('{0}{1}_{2}.json'.format(path, model.name, optimization.optimizationSetting), 'w') as outFile:
                            outFile.write(jsonpickle.encode({'model': model, 'datasets': datasets, 'bestGuesses': optimization.bestGuesses, 'optimizationSetting': optimization.optimizationSetting, 'energyFunction': self.energyFunction}))

                # to store the paths to the optimization results files.
                with open(self.path + '/{0}_{1}.txt'.format(optimizationParameters['optimizationSetting'], self.energyFunction), 'w') as resultsFile:
                    for dirpath, _, filenames in os.walk(path):
                        for f in filenames:
                            if f.endswith('{0}.json'.format(optimizationParameters['optimizationSetting'])):
                                resultsFile.write(os.path.abspath(os.path.join(dirpath, f)) + '\n')

    def _initLogging(self):

        """
        To create logger to store optimization process information.
        :return: logging
        """

        formatter = logging.Formatter('%(asctime)s [%(levelname)s] %(module)s: %(message)s', datefmt='%m/%d/%Y %H:%M:%S' )
        fileHandler = logging.FileHandler(self.path + '/logFile.log')
        fileHandler.setFormatter(formatter)
        logging.getLogger().addHandler(fileHandler)
        logging.getLogger().setLevel(logging.INFO)
        return logging



#!/usr/bin/python3
# -*- coding: utf-8 _*_

"""
Routines for working with moiety modeling.

This package includes the following modules:

``modeling``
    This module provides the :class:`~moiety_modeling.modeling.Dataset` class to organize a single mass spectroscopy
    profile dataset into a dictionary-based data structure, the :class:`~moiety_modeling.modeling.ModelOptimization`
    and derived classes for performing a single model optimization, and the :class:`~moiety_modeling.modeling.OptimizationManager`
    class to manage the optimization process of moiety modeling.

``model``
    This module provides the :class:`~moiety_modeling.model.Moiety`, :class:`~moiety_modeling.model.Molecule`,
    :class:`~moiety_modeling.model.Relationship`, and :class:`~moiety_modeling.model.Model` classes for representing a
    moiety model.

``analysis``
    This module provides several classes to analyze the optimization results, select the optimal model, and visualize the
    results. The :class:`~moiety_modeling.analysis.ResultAnalysis` class is responsible for generating general statistics
    from the optimization results. The :class:`~moiety_modeling.analysis.ModelRank` class selects the model that best
    reflects the observed isotopologue profile. The :class:`~moiety_modeling.analysis.ComparisonTable` class compares
    the optimal model selected under different optimization settings. The :class:`~moiety_modeling.analysis.PlotMoietyDistribution`
    class plots the distribution of moiety value of the moiety model. The :class:`~moiety_modeling.analysis.PlotIsotopologueIntensity`
    class plots comparison of the observed and the calculated isotopologue intensity.

``cli``
    This module provides a command-line interface for the ``moiety_modeling`` package.

"""

__version__ = "1.0.2"

from .modeling import *
from .model import *




#!/usr/bin/python3

"""
moiety_modeling.model
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module provides the :class:`~moiety_modeling.model.Moiety` class, the :class:`~moiety_modeling.model.Molecule`
class, the :class:`~moiety_modeling.model.Relationship` class, and the :class:`~moiety_modeling.model.Model`
class to construct moiety model.

"""

import sys
import itertools
import collections

__all__ = ['Moiety', 'Relationship', 'Molecule', 'Model']


class Moiety:

    """Moiety class describes the :class:`~moiety_modeling.model.Moiety` entity in the moiety model."""

    def __init__(self, name, maxIsotopeNum, isotopeStates, nickname=' ', ranking=0):
        """Moiety initializer.

        :param str name: the name of the moiety.
        :param dict maxIsotopeNum: the dictionary of the labeling isotopes and the corresponding number in the moiety. eg: {'13C': 5, '15N': 3}
        :param dict isotopeStates: the dictionary of the labeling isotopes and the corresponding states in the moiety. eg: {'13C': [0, 2, 5], '15N': [2, 3]}
        :param str nickname: the nickname of the moiety.
        :param int ranking: the ranking of the moiety.
        """
        self.name = name
        self.maxIsotopeNum = dict(maxIsotopeNum)
        self.isotopeStates = dict(isotopeStates)
        self.nickname = nickname
        self.ranking = ranking
        self.states = self._createStates()

    def _createStates(self):

        """
        To calculate the states of moiety. The combinations of each isotope and its state.
        :return: the list of states of the moiety. eg: ['13C0.15N2', '13C2.15N3']
        """

        states = []
        isotopes = sorted(self.isotopeStates)
        for moietyStateComposition in itertools.product(*[self.isotopeStates[isotope] for isotope in isotopes]):
            states.append('.'.join([isotopes[i] + str(moietyStateComposition[i]) for i in range(len(isotopes))]))
        return states

    def __str__(self):

        """Convert the moiety object into string representation.
        :return: the string representation.
        """

        isotopes = sorted(self.isotopeStates)
        string = "Moiety: {0}({1}) \tIsotopes: {2} \tIsotopeStates: ".format(self.name, self.nickname, ','.join(["{0}({1})".format(isotope, self.maxIsotopeNum[isotope]) for isotope in isotopes]))
        for isotope in isotopes:
            string += "{0}[{1}]  ".format(isotope, ', '.join(str(i) for i in self.isotopeStates[isotope]))
        string += "\nMoietyStates: [{0}]".format(', '.join(self.states))
        return string


class Relationship:

    """Relationship class describes the relationship between moiety states in the moiety model."""

    def __init__(self, moiety, moietyState, equivalentMoiety, equivalentMoietyState, operator=None, coefficient=None):
        """Relationship initializer.

        :param moiety: the :class:`~moiety_modeling.model.Moiety` in the relationship.
        :type moiety: :class:`~moiety_modeling.model.Moiety`
        :param str moietyState: the state of the moiety (eg: '13C0.15N1').
        :param equivalentMoiety: :class:`~moiety_modeling.model.Moiety` in the relationship.
        :type equivalentMoiety: :class:`~moiety_modeling.model.Moiety`
        :param str equivalentMoietyState: the state of the equivalentMoiety (eg: '13C0.15N1').
        :param str operator: the operator of the relationship ('*', '/').
        :param double coefficient: the coefficient of the relationship.
        """

        self.moiety = moiety
        self.moietyState = moietyState
        self.equivalentMoiety = equivalentMoiety
        self.equivalentMoietyState = equivalentMoietyState
        self.varName = '{0}[{1}]'.format(self.moiety.name, self.moietyState)
        self.equivalentVarName = '{0}[{1}]'.format(self.equivalentMoiety.name, self.equivalentMoietyState)

        if operator == "/" and coefficient == 0:
            sys.exit("Division by zero in relationship")
        if operator == "*":
            self.multiplier = coefficient
        elif operator == "/":
            self.multiplier = 1/coefficient
        else:
            self.multiplier = 1

    def __str__(self):

        """Converts the relationship to a string representation.
        :return: the string representation.
        """

        if self.multiplier == 1:
            return "Relationship: {0} = {1}".format(self.varName, self.equivalentVarName)
        else:
            return "Relationship: {0} = {1} * {2}".format(self.varName, self.equivalentVarName, self.multiplier)


class Molecule:

    """Molecule class describes the :class:`~moiety_modeling.model.Molecule` entity in the moiety model"""

    def __init__(self, name, moieties):
        """Molecule initializer.

        :param str name: the name of the molecule.
        :param list moieties: the list of :class:`~moiety_modeling.model.Moiety` instances that make up for the :class:`~moiety_modeling.model.Molecule`.
        """

        self.name = name
        self.moieties = list(moieties)
        self.allStates = self._createAllStates()
        self.standardStates = self._createStandardStates()

    def __eq__(self, other):

        return self.name == other.name

    def _createAllStates(self):

        """Calculate all the possible molecule states based on the max isotope number.

        :return: the list of the molecule states. eg: ['13C0.15N0', '13C1.15N0'...]
        """
        # moiety1: {'13C': 6, '15N': 3}, {'13C': [0, 2, 5], '15N': [2, 3]}; moiety2: {'13C': 4, '15N': 5}, {'13C': [1, 3, 4], '15N': [1, 5]}
        # moietyIsotopeNum = [[6, 4], [3, 5]]
        allStates = []
        isotopes = sorted(self.moieties[0].isotopeStates)
        moietyIsotopeNumList = [[moiety.maxIsotopeNum[isotope] for moiety in self.moieties] for isotope in sorted(self.moieties[0].isotopeStates)]
        allStateCompositions = [[i for i in range(sum(isotopeNumList)+1)] for isotopeNumList in moietyIsotopeNumList]
        for moleculeStateComposition in itertools.product(*allStateCompositions):
            allStates.append('.'.join([isotopes[i]+str(moleculeStateComposition[i]) for i in range(len(isotopes))]))
        return allStates

    def _createStandardStates(self):

        """ To calculate the standard molecule states based on the states of each moiety in the molecule.

        :return: the standard molecule states.
        """

        # moietyStates : [[[0, 2, 5], [1, 3, 4]] , [[2, 3], [1, 5]]]
        moleculeStates = collections.defaultdict(list)
        isotopes = sorted(self.moieties[0].isotopeStates)
        moietyStates = [[moiety.isotopeStates[isotope] for moiety in self.moieties] for isotope in sorted(self.moieties[0].isotopeStates)]
        for moleculeStateCompositions in itertools.product(*[itertools.product(*i) for i in moietyStates]):
            isotopeNum = [sum(i) for i in moleculeStateCompositions]
            moleculeComposition = [[] for i in range(len(self.moieties))]
            # example of moleculeStateComposition: [[0 (moiety1), 1 (moiety2)] (13C), [2, 1] (15N))
            for i in range(len(moleculeStateCompositions)):
                for j in range(len(moleculeComposition)):
                    moleculeComposition[j].append(isotopes[i]+str(moleculeStateCompositions[i][j]))
            # example of molecule isotopomer: [g[13C0.15N2], u[13C1.15N1]]
            isotopomer = ['{0}[{1}]'.format(self.moieties[i].name, '.'.join(moleculeComposition[i])) for i in range(len(self.moieties))]
            isotopologue = '.'.join(['{0}{1}'.format(isotopes[i], str(isotopeNum[i])) for i in range(len(isotopeNum))])
            moleculeStates[isotopologue].append(isotopomer)
        return moleculeStates

    def __str__(self):

        """Converts the molecule object into string representation.

        :return: The string representation of the molecule.
        """

        string = "Molecule: {0}\t moieties: {1}\n".format(self.name, ','.join([moiety.name for moiety in self.moieties]))
        for isotopologue in self.standardStates:
            string += " calc[{0}] = {1}\n".format(isotopologue, ' + '.join(['.'.join(isotopomer) for isotopomer in self.standardStates[isotopologue]]))
        return string


class Model:

    """
    Model class describes the moiety :class:`~moiety_modeling.model.Model` entity.
    """

    def __init__(self, name, moieties, molecules, relationships=None):
        """Model initializer.

        :param str name: the name of the model.
        :param list moieties: a list of :class:`~moiety_modeling.model.Moiety` instances.
        :param list molecules: a list of :class:`~moiety_modeling.model.Molecule` instances.
        :param list relationships: a list of :class:`~moiety_modeling.model.Relationship` instances.
        """
        self.name = name
        self.molecules = list(molecules)
        self.moieties = list(moieties)
        self.relationships = [] if relationships is None else list(relationships)
        self.orderedMoieties = []
        self.properties = {}
        self.parameterNum = 0
        self._buildModel()

    def _buildModel(self):

        """To build the model properties.

        :return:
        """

        # Check for circular dependencies in relationships
        relationshipDict = {relationship.varName: relationship for relationship in self.relationships}
        for relationship in self.relationships:
            foundVarNames = [relationship.varName]
            testRelationship = relationship
            moietiesVisited = collections.defaultdict(lambda: 0)
            moietiesVisited[relationship.moiety.name] = 1
            while testRelationship and testRelationship.equivalentVarName not in foundVarNames:
                moietiesVisited[testRelationship.equivalentMoiety.name] += 1
                foundVarNames.append(testRelationship.equivalentVarName)
                testRelationship = relationshipDict[testRelationship.equivalentVarName] if testRelationship.equivalentVarName in relationshipDict else None
            if testRelationship:
                # This tests the relationship between the different moieties. There is no circle between moieties. Like g[5] = r[2], r[2] = a[3], a[3] = g[5];
                sys.exit("Circular relationship starting with {0} and ending with {1}".format(relationship.varName, testRelationship.varName))
            if [ value for value in moietiesVisited.values() if value > 2 ]:
                # This test the relationship within the same moiety. Only one relationship between the moiety is allowed. Like g[5] = g[3], g[3] = g[2], g[2] = g[1]; why?
                sys.exit("Relayed Intra moiety relationship starting with {0}".format(relationship.varName))

        # Determine moiety relationships
        # This can cause the problems of ranking, so if there are several relationships between two moieties, we try to put one moiety at one side, another moiety at another side of the equation
        # rather than put them randomly.
        moietyRelationshipDict = {}
        for relationship in self.relationships:
            if relationship.moiety.name != relationship.equivalentMoiety.name:
                if relationship.moiety.name not in moietyRelationshipDict:
                    moietyRelationshipDict[relationship.moiety.name] = {'moiety': relationship.moiety, 'equivalentMoieties': {relationship.equivalentMoiety}}
                else:
                    moietyRelationshipDict[relationship.moiety.name]['equivalentMoieties'].add(relationship.equivalentMoiety)

        for relationship in moietyRelationshipDict.values():
            foundMoieties = {relationship['moiety']}
            testMoieties = relationship['equivalentMoieties']
            while testMoieties and not foundMoieties & testMoieties:
                newTestMoieties = set()
                for moiety in testMoieties:
                    moiety.ranking += 1 # The higher the moiety ranking is, the more likely the moiety is correlated with other moieties.
                    if moiety.name in moietyRelationshipDict:
                        newTestMoieties.update(moietyRelationshipDict[moiety.name]['equivalentMoieties'] - testMoieties)
                foundMoieties.update(testMoieties)
                testMoieties = newTestMoieties
            if testMoieties:
                sys.exit("Circular moiety relationships starting with {0} and ending with {1}".format(relationship.moiety.name, testMoieties[0].name))

        self.orderedMoieties = list(self.moieties)
        self.orderedMoieties.sort(key=lambda moiety: moiety.ranking, reverse=True)

        for moiety in self.orderedMoieties:
            numOfParameters = len(moiety.states) - 1
            preAssignedStates = []
            calculatedStates = []
            postAssignedStates = []
            elements = []
            elementNames = []
            divisors = {state: 1 for state in moiety.states}

            for state in moiety.states:
                varName = '{0}[{1}]'.format(moiety.name, state)
                if varName in relationshipDict and relationshipDict[varName].equivalentMoiety in self.orderedMoieties:
                    numOfParameters -= 1
                    if relationshipDict[varName].moiety.name == relationshipDict[varName].equivalentMoiety.name:
                        # the relationship is appended to the postAssignedStates and preAssignedStates
                        postAssignedStates.append(relationshipDict[varName])
                        divisors[relationshipDict[varName].equivalentMoietyState] += relationshipDict[varName].multiplier
                    else:
                        preAssignedStates.append(relationshipDict[varName])
            for state in moiety.states:
                varName = '{0}[{1}]'.format(moiety.name, state)
                if varName not in relationshipDict or relationshipDict[varName].equivalentMoiety not in self.orderedMoieties:
                    # the states of moiety and its divisor are attached to the calculatedStates.
                    calculatedStates.append([varName, divisors[state]])
            for x in range(numOfParameters):
                self.parameterNum += 1
                elementNames.append('{0}.p{1}'.format(moiety.name, x))
                elements.append(self.parameterNum-1)
            self.properties[moiety.name] = {'preAssignedStates': preAssignedStates, 'postAssignedStates': postAssignedStates, 'calculatedStates': calculatedStates, 'elements': elements, 'elementNames' : elementNames}

    def calculateMoietyState(self, vector):

        """To calculate the value of the all moietyStates.

        :param vector: the list of parameters of the model.
        :param dataset: dataset object.
        :return: the dictionary containing the value of the moietyStates.
        """

        conversionDict = {}
        moietyStates = {}
        for moiety in self.orderedMoieties:
            properties = self.properties[moiety.name]
            if properties['preAssignedStates']:
                for relationship in properties['preAssignedStates']:
                    conversionDict[relationship.varName] = conversionDict[relationship.equivalentVarName] * relationship.multiplier
                conversionDict[moiety.name + 'PreAssignedTotal'] = sum([conversionDict[relationship.varName] for relationship in properties['preAssignedStates']])
                if conversionDict[moiety.name + 'PreAssignedTotal'] > 1:
                    for relationship in properties['preAssignedStates']:
                        conversionDict[relationship.varName] /= conversionDict[moiety.name + 'PreAssignedTotal']
                        conversionDict[moiety.name + 'PreAssignedTotal'] = 1
            else:
                conversionDict[moiety.name + 'PreAssignedTotal'] = 0
            calculatedStates = properties['calculatedStates']
            if properties['elements']:
                conversionDict[moiety.name + 'Parameters'] = [max(0, vector[i]) if vector[i] < 1 else 1 for i in properties['elements'] ]
                conversionDict[moiety.name + 'ParameterMax'] = max(conversionDict[moiety.name + 'Parameters'])
                conversionDict[moiety.name + 'ParametersSum'] = sum(conversionDict[moiety.name + 'Parameters'])
                conversionDict[calculatedStates[0][0]] = (1 - conversionDict[moiety.name + 'ParameterMax']) * (1 - conversionDict[moiety.name + 'PreAssignedTotal']) / calculatedStates[0][1]
                if conversionDict[moiety.name + 'ParametersSum'] == 0:
                    conversionDict[moiety.name + 'ParametersSum'] = 1
                for y in range(len(properties['elements'])):
                    conversionDict[calculatedStates[y + 1][0]] = conversionDict[moiety.name + 'Parameters'][y] / conversionDict[moiety.name + 'ParametersSum'] * conversionDict[moiety.name + 'ParameterMax'] * (1 - conversionDict[moiety.name + 'PreAssignedTotal']) / calculatedStates[y + 1][1]
            elif calculatedStates:
                conversionDict[calculatedStates[0][0]] = (1 - conversionDict[moiety.name + 'PreAssignedTotal']) / calculatedStates[0][1]
            for relationship in properties['postAssignedStates']:
                conversionDict[relationship.varName] = conversionDict[relationship.equivalentVarName] * relationship.multiplier
            for state in moiety.states:
                moietyState = '{0}[{1}]'.format(moiety.name, state)
                moietyStates[moietyState] = conversionDict[moietyState]
        return moietyStates

    def __str__(self):

        """
        String representation of the model.
        :return: string.
        """

        string = "\n#####Conversion and calculation script for model {0}.\n".format(self.name)
        for moiety in self.orderedMoieties:
            properties = self.properties[moiety.name]
            string += "# {0} conversion\n".format(moiety.name)
            if properties['preAssignedStates']:
                for relationship in properties['preAssignedStates']:
                    string += " {0}[{1}] = {2}[{3}]".format(moiety.name, relationship.moietyState, relationship.equivalentMoiety.name, relationship.equivalentMoietyState)
                    string += "" if relationship.multiplier == 1 else " * {0}".format(relationship.multiplier)
                    string += "\n"
                string += " {0}PreAssignedTotal = {1}\n\n".format(moiety.name, " + ".join(["{0}[{1}]".format(moiety.name, relationship.moietyState) for relationship in properties['preAssignedStates']]))
                string += " if {0}PreAssignedTotal > 1:\n".format(moiety.name)
                for relationship in properties['preAssignedStates']:
                    string += "     {0}[{1}] /= {2}PreAssignedTotal\n".format(moiety.name, relationship.moietyState, moiety.name)
                string += "     {0}PreAssignedTotal = 1\n".format(moiety.name)
            else:
                string += " {0}PreAssignedTotal = 0\n".format(moiety.name)
            calculatedStates = properties['calculatedStates']
            if properties['elements']:
                string += " {0}Parameters = [vector[i] for i in {1}]\n".format(moiety.name, properties['elements'])
                string += " {0}ParameterMax = max({1}Parameters)\n".format(moiety.name, moiety.name)
                string += " {0}ParametersSum = 0\n".format(moiety.name)
                string += " for parameter in {0}Parameters:\n".format(moiety.name)
                string += "     {0}ParametersSum += parameter\n".format(moiety.name)
                string += " {0}[{1}] = (1 - {2}ParameterMax) * (1 - {3}PreAssignedTotal)".format(moiety.name, calculatedStates[0][0], moiety.name, moiety.name)
                string += "" if calculatedStates[0][1] == 1 else " / {0}".format(calculatedStates[0][1])
                string += "\n"
                for y in range(len(properties['elements'])):
                    string += " {0}[{1}] = {2}Parameters[{3}] / {4}ParametersSum * {5}ParameterMax * (1 - {6}PreAssignedTotal)".format(moiety.name, calculatedStates[y+1][0], moiety.name, y, moiety.name, moiety.name, moiety.name)
                    string += "" if calculatedStates[y+1][1] == 1 else " / {0}".format(calculatedStates[y+1][1])
                    string += "\n"
            elif calculatedStates:
                string += " {0}[{1}] = (1 - {2}PreAssignedTotal)".format(moiety.name, calculatedStates[0][0], moiety.name)
                string += "" if calculatedStates[0][1] == 1 else " / {0}".format(calculatedStates[0][1])
                string += "\n"
            for relationship in properties['postAssignedStates']:
                string += " {0}[{1}] = {2}[{3}]".format(moiety.name, relationship.moietyState, moiety.name, relationship.equivalentMoietyState)
                string += "" if relationship.multiplier == 1 else " * {0}".format(relationship.multiplier)
                string += "\n"
            string += "\n"
        string += " conversion = {}\n"
        for moiety in self.orderedMoieties:
            string += " conversion['{0}'] = [{1}]\n".format(moiety.name, ','.join(["{0}[{1}]".format(moiety.name, state) for state in moiety.states]))
        return string






#!/usr/bin/python3

"""
The moiety_modeling command-line interface option
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Usage:
    moiety_modeling -h | --help
    moiety_modeling --version
    moiety_modeling modeling [--combinedData=<combined_jsonfile>] [--models=<models_jsonfile] [--datasets=<datasets_jsonfile>] [--optimizations=<optimizations_jsonfile>] [--working=<working_dir>] [--repetition=<optim_count>] [--split] [--force] [--multiprocess] [--energyFunction=<function>] [--printOptimizationScripts]
    moiety_modeling analyze optimizations --a <optimizationPaths_txtfile>  [--working=<working_dir>]
    moiety_modeling analyze optimizations --s <optimizationResults_jsonfile> [--working=<working_dir>]
    moiety_modeling analyze rank <analysisPaths_txtfile> [--working=<working_dir>] [--rankCriteria=<rankCriteria>]
    moiety_modeling analyze table <rankPaths_txtfile> [--working=<working_dir>]
    moiety_modeling plot moiety <analysisResults_jsonfile> [--working=<working_dir>]
    moiety_modeling plot isotopologue <analysisResults_jsonfile> [--working=<working_dir>]

Options:
    -h, --help                                       Show this screen.
    --version                                        Show version.
    --combinedData=<combined_jsonfile>               JSON description file of the combined data (eg: models, datasets, optimization settings)
    --models=<models_jsonfile>                       JSON description file of the moiety models.
    --datasets=<datasets_jsonfile>                   JSON description file of the datasets.
    --optimizations=<optimizations_jsonfile>         JSON description file of the optimization setting.
    --working=<working_dir>                          Alternative path to save the results.
    --repetition=<optim_count>                       The number of optimization repetitions to perform [default: 100].
    --split                                          To split the datasets or not.
    --force                                          To force optimization process if error occurs.
    --multiprocess                                   To perform with multiprocessing or not.
    --printOptimizationScripts                       To print the optimization script or not.
    --a                                              To analyze a bunch of optimization results together with the path file.
    --s                                              To analyze a single moiety model optimization results.
    --energyFunction=<enegyFunction>                 The energyFunction of the moiety modeling optimization.
    --optimizationSetting=<optimizationSetting>      The optimization setting of the moiety modeling optimization.
    --rankCriteria=<rankCriteria>                    The criteria for model ranking [default: AICc].
"""

import jsonpickle
import os
from . import modeling
from . import analysis


def cli(args):

    if args['modeling']:

        dirName = ''

        combinedData = {}
        if args['--combinedData']:
            dirName += os.path.splitext(os.path.basename(args['--combinedData']))[0] + '_'
            with open(args['--combinedData']) as combinedFile:
                combinedData = jsonpickle.decode(combinedFile.read())

        if args['--models']:
            dirName += os.path.splitext(os.path.basename(args['--models']))[0] + '_'
            with open(args['--models']) as modelFile:
                models = jsonpickle.decode(modelFile.read())['models']
        elif 'models' in combinedData:
            models = combinedData['models']
        else:
            raise ImportError("Model is missing!")

        if args['--datasets']:
            dirName += os.path.splitext(os.path.basename(args['--datasets']))[0] + '_'
            with open(args['--datasets']) as datasetFile:
                datasets = jsonpickle.decode(datasetFile.read())['datasets']
        elif 'datasets' in combinedData:
            datasets = combinedData['datasets']
        else:
            raise ImportError("Dataset is missing!")

        if args['--optimizations']:
            dirName += os.path.splitext(os.path.basename(args['--optimizations']))[0] + '_'
            with open(args['--optimizations']) as settingFile:
                optimizations = jsonpickle.decode(settingFile.read())['optimizations']
        elif 'optimizations' in combinedData:
            optimizations = combinedData['optimizations']
        else:
            raise ImportError("Optimization setting is missing!")

        times = int(args['--repetition']) if args['--repetition'] else 100
        dirName += 'T_{0}_'.format(times)

        dirName += 'S_' if args['--split'] else 'noS_'

        dirName += 'M_' if args['--multiprocess'] else 'noM_'

        energyFunction = args['--energyFunction'] if args['--energyFunction'] else 'logDifference'
        dirName += energyFunction

        if args['--working']:
            os.mkdir(args['--working'] + '/' + dirName)
            path = args['--working'] + '/' + dirName

        else:
            if args['--datasets']:
                datasetPath = os.path.dirname(args['--datasets'])
                os.mkdir(datasetPath + '/' + dirName)
                path = datasetPath + '/' + dirName

            else:
                combinedPath = os.path.dirname(args['--combinedData'])
                os.mkdir(combinedPath + '/' + dirName)
                path = combinedData + '/' + dirName

        manager = modeling.OptimizationManager(models, datasets, optimizations, path, args['--split'], args['--multiprocess'], args['--force'], times, energyFunction, args['--printOptimizationScripts'])
        manager.optimizeModels()

    elif args['analyze'] or args['plot']:

        if args['optimizations']:
            if args['--a']:
                base = os.path.basename(args['<optimizationPaths_txtfile>'])
                baseName = os.path.splitext(base)[0]
                path = args['--working'] if args['--working'] else '{0}/{1}_analysis/'.format(os.path.dirname(args['<optimizationPaths_txtfile>']), baseName)
                os.mkdir(path)
                analysisPathFile = open('{0}{1}_analysis.txt'.format(path, baseName), 'w')
                with open(args['<optimizationPaths_txtfile>']) as pathFile:
                    for fileName in pathFile:
                        resAnalysis = analysis.ResultsAnalysis(fileName.replace("\n", ""), path)
                        resAnalysis.analyze()
                        analysisPathFile.write(resAnalysis.jsonfilePath + '\n')
                analysisPathFile.close()
            elif args['--s']:
                path = args['--working'] if args['--working'] else None
                resAnalysis = analysis.ResultsAnalysis(args['<optimizationResults_jsonfile>'], path)
                resAnalysis.analyze()
            else:
                raise ImportError("Something is missing for optimization analysis!")

        elif args['rank']:
            if args['<analysisPaths_txtfile>']:
                path = args['--working'] if args['--working'] else None
                selectionCriteria = args['--rankCriteria'] if args['--rankCriteria'] else 'AICc'
                modelRank = analysis.ModelRank(args['<analysisPaths_txtfile>'], path, selectionCriteria)
                modelRank.rank()
            else:
                raise ImportError("Result path file is missing for model rank!")

        elif args['table']:
            if args['<rankPaths_txtfile>']:
                path = args['--working'] if args['--working'] else None
                comparisonTable = analysis.ComparisonTable(args['<rankPaths_txtfile>'], path)
                comparisonTable.makeTable()
            else:
                raise ImportError("Rank path file is missing for comparison table!")

        elif args['moiety']:
            if args['<analysisResults_jsonfile>']:
                path = args['--working'] if args['--working'] else None
                moietyPlot = analysis.PlotMoietyDistribution(args['<analysisResults_jsonfile>'], path)
                moietyPlot.plotMoiety()
            else:
                ImportError("Analysis results jsonfile is missing for moiety distribution plot!")

        elif args['isotopologue']:
            if args['<analysisResults_jsonfile>']:
                path = args['--working'] if args['--working'] else None
                isotopologuePlot = analysis.PlotIsotopologueIntensity(args['<analysisResults_jsonfile>'], path)
                isotopologuePlot.plotIsotopologue()
            else:
                ImportError("Analysis results jsonfile is missing for observed and calculated isotoplogue comparison!")











#!/usr/bin/python3

import re
from setuptools import setup, find_packages


def find_version():
    with open('moiety_modeling/__init__.py', 'r') as fd:
        version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                            fd.read(), re.MULTILINE).group(1)
    if not version:
        raise RuntimeError('Cannot find version information')
    return version

REQUIRES = [
    "docopt",
    "jsonpickle",
    "matplotlib",
    "numpy",
    "SAGA-optimize",
    "scipy"
]

setup(

    name='moiety_modeling',
    version=find_version(),
    packages=find_packages(),
    license="The Clear BSD License",
    author="Huan Jin",
    author_email="hji236@g.uky.edu",
    description="Moiety Modeling Implementation",
    keywords="moiety modeling model optimization model selection",
    url="https://github.com/MoseleyBioinformaticsLab/moiety_modeling.git",
    install_requires=REQUIRES,
    long_description=open('README.rst').read(),
    platforms='any',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',

    ],


)